@@ -233,6 +233,8 @@ static struct agent *agent = NULL;
 struct pending_request {
 	struct server *server;
 	gchar address[18];
+	gchar *adapter_path;
+	guint watch;
 	gint nsk;
 };
 
@@ -833,6 +835,9 @@ static void agent_reply(DBusPendingCall *call, gpointer user_data)
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	const gchar *name;
 	DBusError derr;
+	gboolean *got_reply = user_data;
+
+	*got_reply = TRUE;
 
 	/* Received a reply after the agent exited */
 	if (!agent)
@@ -854,6 +859,7 @@ static void agent_reply(DBusPendingCall *call, gpointer user_data)
 				DBUS_TYPE_INVALID)) {
 		/* Splits folder and name */
 		const gchar *slash = strrchr(name, '/');
+		debug("Agent replied with %s", name);
 		if (!slash) {
 			agent->new_name = g_strdup(name);
 			agent->new_folder = NULL;
@@ -869,17 +875,8 @@ static void agent_reply(DBusPendingCall *call, gpointer user_data)
 static gboolean auth_error(GIOChannel *io, GIOCondition cond,
 			gpointer user_data)
 {
-	DBusMessage *msg;
-
 	agent->auth_pending = FALSE;
 
-	msg = dbus_message_new_method_call(agent->bus_name, agent->path,
-					"org.openobex.Agent", "Cancel");
-
-	dbus_connection_send_with_reply(connection, msg, NULL, -1);
-
-	dbus_message_unref(msg);
-
 	return FALSE;
 }
 
@@ -896,6 +893,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	const gchar *bda = address;
 	gchar *path;
 	guint watch;
+	gboolean got_reply;
 
 	if (!agent)
 		return -1;
@@ -910,7 +908,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	addrlen = sizeof(addr);
 
 	if (getpeername(fd, (struct sockaddr *) &addr, &addrlen) < 0)
-		return -1;
+		return -errno;
 
 	ba2str(&addr.rc_bdaddr, address);
 
@@ -939,6 +937,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	dbus_message_unref(msg);
 
 	agent->auth_pending = TRUE;
+	got_reply = FALSE;
 
 	/* Catches errors before authorization response comes */
 	io = g_io_channel_unix_new(fd);
@@ -947,7 +946,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 			auth_error, NULL, NULL);
 	g_io_channel_unref(io);
 
-	dbus_pending_call_set_notify(call, agent_reply, NULL, NULL);
+	dbus_pending_call_set_notify(call, agent_reply, &got_reply, NULL);
 
 	/* Workaround: process events while agent doesn't reply */
 	while (agent && agent->auth_pending)
@@ -955,7 +954,15 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 
 	g_source_remove(watch);
 
-	dbus_pending_call_cancel(call);
+	if (!got_reply) {
+		dbus_pending_call_cancel(call);
+		msg = dbus_message_new_method_call(agent->bus_name,
+							agent->path,
+							"org.openobex.Agent",
+							"Cancel");
+		g_dbus_send_message(connection, msg);
+	}
+
 	dbus_pending_call_unref(call);
 
 	if (!agent || !agent->new_name)
@@ -990,6 +997,9 @@ static void service_reply(DBusPendingCall *call, gpointer user_data)
 		close(pending->nsk);
 
 done:
+	if (pending->watch)
+		g_source_remove(pending->watch);
+	g_free(pending->adapter_path);
 	g_free(pending);
 	dbus_message_unref(reply);
 }
@@ -997,17 +1007,17 @@ done:
 static gboolean service_error(GIOChannel *io, GIOCondition cond,
 			gpointer user_data)
 {
+	struct pending_request *pending = user_data;
 	DBusMessage *msg;
 
-	if (any->path == NULL)
-		return FALSE;
+	pending->watch = 0;
 
 	msg = dbus_message_new_method_call("org.bluez",
-					any->path,
+					pending->adapter_path,
 					"org.bluez.Service",
 					"CancelAuthorization");
 
-	dbus_connection_send_with_reply(system_conn, msg, NULL, -1);
+	dbus_connection_send(system_conn, msg, NULL);
 
 	dbus_message_unref(msg);
 
@@ -1021,7 +1031,6 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 	DBusMessage *msg;
 	DBusPendingCall *pcall;
 	GIOChannel *io;
-	guint watch;
 	const char *paddr = pending->address;
 	const char *path;
 	DBusError derr;
@@ -1031,7 +1040,7 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 		error("Replied with an error: %s, %s",
 				derr.name, derr.message);
 		dbus_error_free(&derr);
-		goto done;
+		goto failed;
 	}
 
 	dbus_message_get_args(reply, NULL,
@@ -1039,6 +1048,8 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 			DBUS_TYPE_INVALID);
 
 	debug("FindAdapter -> %s", path);
+	pending->adapter_path = g_strdup(path);
+	dbus_message_unref(reply);
 
 	msg = dbus_message_new_method_call("org.bluez", path,
 			"org.bluez.Service", "RequestAuthorization");
@@ -1050,30 +1061,31 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 	if (!dbus_connection_send_with_reply(system_conn,
 					msg, &pcall, TIMEOUT)) {
 		dbus_message_unref(msg);
-		goto done;
+		goto failed;
 	}
 
 	dbus_message_unref(msg);
 
 	debug("RequestAuthorization(%s, %x)", paddr, pending->server->handle);
 
-	if (!dbus_pending_call_set_notify(pcall, service_reply, pending, NULL)) {
-		close(pending->nsk);
-		g_free(pending);
-		goto done;
-	}
+	if (!dbus_pending_call_set_notify(pcall, service_reply, pending, NULL))
+		goto failed;
 
 	dbus_pending_call_unref(pcall);
 
 	/* Catches errors before authorization response comes */
 	io = g_io_channel_unix_new(pending->nsk);
-	watch = g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
-			G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-			service_error, NULL, NULL);
+	pending->watch = g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
+					G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+					service_error, pending, NULL);
 	g_io_channel_unref(io);
 
-done:
-	dbus_message_unref(reply);
+	return;
+
+failed:
+	g_free(pending->adapter_path);
+	close(pending->nsk);
+	g_free(pending);
 }
 
 gint request_service_authorization(struct server *server, gint nsk)
